from collections import namedtuple
from loguru import logger

class Defaults():
    timeZone: str = 'UTC'
    useDateTimeOffset: bool = True
    copyGraphQLString: bool = False

class Structure():
    queryStructure = f'''query inventories {{
    inventories (pageSize:1000) {{
        name
        inventoryId
        isDomainUserType
        hasValidityPeriods
        historyEnabled
        propertyUniqueness {{
            uniqueKey
            properties
            }}
        variant {{
            name
            properties {{
                name
                type
                isArray
                nullable
                }}
            }}
        properties {{
            name
            ...Scalar
            type
            isArray
            nullable
            propertyId
            ... Reference 
            }}
        }}
    }}
    fragment Scalar on IScalarProperty {{
        dataType
        }}
    fragment Reference on IReferenceProperty {{
        inventoryId
        inventoryName
        }}
    '''

    def _introspectionQueryString():
        introspectionQueryString = r'''
            query IntrospectionQuery { __schema { queryType { name } mutationType 
                { name } subscriptionType { name } types { ...FullType } directives
                { name description locations args { ...InputValue } } } }

            fragment FullType on __Type { kind name description fields(includeDeprecated: true) { name description args 
                { ...InputValue } type { ...TypeRef } isDeprecated deprecationReason } inputFields { ...InputValue } interfaces 
                { ...TypeRef } enumValues(includeDeprecated: true) { name  } possibleTypes { ...TypeRef } } 

            fragment InputValue on __InputValue { name description type { ...TypeRef } defaultValue } 
            fragment TypeRef on __Type { kind name ofType { kind name ofType { kind name ofType 
                    { kind name ofType { kind name ofType { kind name ofType { kind name ofType { kind name } } } } } } } }
        '''
        return introspectionQueryString

    def _fullStructureDict(structure) -> dict:
        """Converts the query of all inventories with all fields into a pure dict"""

        def subdict(inputObject, name):
            itemDict = {}
            for item in inputObject:
                itemDict.setdefault(item[name], {})
                for k, v in item.items():
                    itemDict[item[name]].setdefault(k,v)
            return itemDict

        structureDict = {}
        for inventory in structure['inventories']:
            inventoryName = inventory['name']
            structureDict.setdefault(inventoryName, {})
            for definitionKey, definitionValue in inventory.items():
                if not isinstance(definitionValue, (list, dict)):
                    structureDict[inventoryName].setdefault(definitionKey, definitionValue)
                else:
                    if definitionKey == 'properties':
                        subDict = subdict(inventory[definitionKey], 'name')
                        structureDict[inventoryName].setdefault(definitionKey, subDict)
                    if definitionKey == 'propertyUniqueness':
                        subDict = subdict(inventory[definitionKey], 'uniqueKey')
                        structureDict[inventoryName].setdefault(definitionKey, subDict)
                    if definitionKey == 'variant':
                        structureDict[inventoryName].setdefault(definitionKey, {})
                        structureDict[inventoryName][definitionKey].setdefault('name', definitionValue['name'])
                        subDict = subdict(inventory[definitionKey]['properties'], 'name')
                        structureDict[inventoryName][definitionKey].setdefault('properties', subDict)
        return structureDict

    def _fullStructureNT(structure:dict) -> namedtuple:
        """
        Provides the complete data structure of dynamic objects as named tuple. 
        Needs structureDict first
        """
        def _subItem(object:dict):
            Item = namedtuple('Item', object.keys())
            itemDict = {}
            for key, value in object.items():
                if isinstance(value, dict):
                    subItem = _subItem(value)
                    itemDict.setdefault(key, subItem)
                else:
                    itemDict.setdefault(key, value)
            item = Item(**itemDict)
            return item

        Item = namedtuple('Item', structure.keys())
        itemDict = {}
        for key, value in structure.items():
            if isinstance(value, dict):
                subItem = _subItem(value)
                itemDict.setdefault(key, subItem)
            else:
                itemDict.setdefault(key, value)
        return Item(**itemDict)

    def _inventoryNT(structure) -> namedtuple:
        """
        Provides a simplified namedtuple of dynamic objects for interactive usage
        """
        inventoryDict = {key:key for key in structure.keys()}
        Inventories = namedtuple('Inventories', inventoryDict.keys())
        return Inventories(**inventoryDict)

    def _inventoryPropertyNT(structure) -> namedtuple:
        """
        Provides a simplified namedtuple of inventory properties for interactive usage
        """
        Inventory = namedtuple('Inventories', structure.keys())
        inventoryDict = {}

        for inventory in structure.keys():
            propertyDict = {}
            for key in structure[inventory]['properties'].keys():
                propertyDict.setdefault(key, key)
                Properties = namedtuple('Properties', propertyDict.keys())
                properties = Properties(**propertyDict)
            inventoryDict.setdefault(inventory, properties)
        return Inventory(**inventoryDict)


# -*- coding: utf-8 -*-

import string
from typing import Tuple

#: Default Loki emitter version.
emitter_ver: str = "0"
#: Size of LRU cache for LogQL label formatting.
format_label_lru_size: int = 256

#: Success HTTP status code from Loki API.
success_response_code: int = 204

#: Label name indicating logging level.
level_tag: str = "severity"
#: Label name indicating logger name.
logger_tag: str = "logger"

#: String contains chars that can be used in label names in LogQL.
label_allowed_chars: str = "".join((string.ascii_letters, string.digits, "_"))
#: A list of pairs of characters to replace in the label name.
label_replace_with: Tuple[Tuple[str, str], ...] = (
    ("'", ""),
    ('"', ""),
    (" ", "_"),
    (".", "_"),
    ("-", "_"),
)



import base64
from loguru import logger
import os

class AutProgUtils():

    def _upsetFiles(filePaths:list) -> str:
        """Converts a list of files to Base64 provides the upsetFiles graphQL format"""
        
        _files = ''

        for path in filePaths:
            if not os.path.exists(path):
                logger.error(f"Path {path} does not exist.")
                return None
            else: 
                fileName = os.path.split(path)[-1]
                contentBase64 = AutProgUtils._encodeBase64(path)
                _files += f'{{fullName: "{fileName}", contentBase64: "{contentBase64}"}},\n'

        return _files

    def _getFileNames(filePaths:list) -> str:
        """Returns file names from a file path"""
        return [os.path.split(path)[-1] for path in filePaths]


    def _encodeBase64(file:str):
        with open(file, mode='r', encoding='utf-8') as file:
            BOM = '\ufeff'
            content = file.read()
            if content.startswith(BOM):
                content = content[1:]
            content = base64.b64encode(content.encode('utf-8'))
            return content.decode('utf-8')

    def _varsToString(vars:dict, type) -> str:
        """Transforms a dict of function variables to a string. All values will be 
            converted to string.
        """
        if type == 'input': type = 'inputVariables'
        elif type == 'env': type = 'environmentVariables'
        else:
            return ''

        _vars = f'{type}: [\n'
        for key, value in vars.items():
            _vars += f'{{key: "{key}", value: "{str(value)}"}}\n'
        
        _vars += '\n]'

        return _vars

    def _downloadFunctionFile(path, fileName, content):
        with open(path+fileName, 'w') as f:
            f.write(base64.b64decode(content).decode('UTF8'))







from typing import Literal
from uuid import uuid4
import pandas as pd
from loguru import logger

from .utils.ut import Utils

class Authorization:

    def __init__(self, accessToken:str, endpoint:str, client:object) -> None:
        global dyno_client
        dyno_client = client

        self.raiseException = client.raiseException
        self.defaults = client.defaults

        self.accessToken = accessToken
        self.endpoint = endpoint
        self.proxies = client.proxies
        self.structure = client.structure
        self.scheme = client.scheme

    def getVersion(self):
        """
        Returns name and version of the responsible micro service
        """

        return Utils._getServiceVersion(self, 'authorization')

    def _resolve_where(self, where:str):
        # todo: remove function and use Utils. Therefore remove the 'topLevel' thing! Is it coming from dynO?
        resolvedFilter = ''
        if where != None: 
            resolvedFilter = f'({Utils._resolveWhere(self, where)["topLevel"]})'
        
        return resolvedFilter

    def roles(
        self, 
        fields:list=None,
        where:str=None
        ) -> pd.DataFrame:
        """
        Returns a DataFrame of available roles
        """

        key = 'roles'

        if fields != None:
            if type(fields) != list:
                fields = [fields]
            _fields = Utils._queryFields(fields, recursive=True)   
        else:
            _fields =f'''
                id
                name
                revision
            ''' 

        resolvedFilter = self._resolve_where(where)
        graphQLString = f'''query roles {{
            {key}{resolvedFilter}  {{
                {_fields}
            }}
        }}
        '''

        result = Utils._executeGraphQL(self, graphQLString)
        if result == None: return

        df = pd.json_normalize(result[key])
        return df

    def rules(
        self,
        fields:list=None, 
        where:str=None
        ) -> pd.DataFrame:
        """
        Returns a DataFrame of available rules
        """

        key = 'rules'

        if fields != None:
            if type(fields) != list:
                fields = [fields]
            _fields = Utils._queryFields(fields, recursive=True)   
        else:
            _fields =f'''
                id
                filter
                revision
                role {{
                    id
                    name
                }}
                accountReference {{
                    ...on GroupReference {{
                        group {{
                            id
                            name
                        }}
                    }}
                    ...on ServiceAccountReference {{
                        serviceAccount {{
                            id
                            name
                        }}
                    }}
                }}
            ''' 

        resolvedFilter = self._resolve_where(where)
        graphQLString = f'''query Rules {{
            {key}{resolvedFilter}  {{
                {_fields}
            }}
        }}
        '''

        result = Utils._executeGraphQL(self, graphQLString)
        if result == None: return

        df = pd.json_normalize(result[key])
        return df

    def users(
        self,
        fields:list=None, 
        where:str=None) -> pd.DataFrame:
        """
        Returns a DataFrame of available users
        """

        key = 'users'

        if fields != None:
            if type(fields) != list:
                fields = [fields]
            _fields = Utils._queryFields(fields, recursive=True)   
        else:
            _fields =f'''
                id
                name
                username
                providerName
                providerUserId
                revision
            ''' 

        resolvedFilter = self._resolve_where(where)
        graphQLString = f'''query Users {{
            {key}{resolvedFilter}  {{
                {_fields}
            }}
        }}
        '''

        result = Utils._executeGraphQL(self, graphQLString)
        if result == None: return

        df = pd.json_normalize(result[key])
        return df

    def userGroups(
        self,
        fields:list=None, 
        where:str=None
        ) -> pd.DataFrame:
        """
        Returns a DataFrame of available users
        """

        key = 'groups'

        if fields != None:
            if type(fields) != list:
                fields = [fields]
            _fields = Utils._queryFields(fields, recursive=True)   
        else:
            _fields =f'''
                id
                name
            ''' 

        resolvedFilter = self._resolve_where(where)
        graphQLString = f'''query Groups {{
            {key}{resolvedFilter}  {{
                {_fields}
            }}
        }}
        '''

        result = Utils._executeGraphQL(self, graphQLString)
        if result == None: return

        df = pd.json_normalize(result[key])
        return df

    def serviceAccounts(
        self,
        fields:list=None, 
        where:str=None
        ) -> pd.DataFrame:
        """
        Returns a DataFrame of available service accounts.
        """

        key = 'serviceAccounts'

        if fields != None:
            if type(fields) != list:
                fields = [fields]
            _fields = Utils._queryFields(fields, recursive=True)   
        else:
            _fields =f'''
                id
                name
            ''' 

        resolvedFilter = self._resolve_where(where)
        graphQLString = f'''query ServiceAccounts {{
            {key}{resolvedFilter}  {{
                {_fields}
            }}
        }}
        '''

        result = Utils._executeGraphQL(self, graphQLString)
        if result == None: return

        df = pd.json_normalize(result[key])
        return df

    def createRole(
        self,
        inventoryName:str, 
        roleName:str,
        userGroups:list=None, 
        objectPermissions:list=['Create', 'Delete'], 
        propertiesPermissions:list=['Read', 'Update']
        ) -> None:

        """
        Creates a role and sets all rights to all properties

        Parameters:
        ----------
        inventoryName : str
            The name of the inventory for which the new role authorizes rights.
        roleName : str
            Name of the new role.
        userGroup : list = None
            List of user group names. If None, the role will be created without attaching user groups.
        objectPermissions : list = ['Create', 'Delete']
            Default is 'Create' and 'Delete' to allow creating and deleting items of the specified inventory.
            Other entries are not allowed.
        propertiesPermissions : list = ['Read', 'Update']
            Default is 'Read' and 'Update'. All properties will receive 
            the specified rights. Other entries are not allowed.
            Permissions are not extended on referenced inventories!
        """

        correlationId = str(uuid4())
        with logger.contextualize(correlation_id=correlationId):

            # Parameter validation
            try:
                dyno_client.structure[inventoryName]
            except:
                Utils._error(self, f"Unknown inventory '{inventoryName}'")
                return
            
            try:
                roles = self.roles()
                if roleName in list(roles['name']):
                    Utils._error(self, f"Role '{roleName}' already exists.")
                    return
            except:
                pass

            if isinstance(userGroups, str):
                userGroups = [userGroups]

            if userGroups != None:
                # 'in' is not supported therefore load all groups
                dfUserGroups = self.userGroups()
                falseUserGroups = []
                for group in userGroups:
                    if group not in list(dfUserGroups['name']):
                        falseUserGroups.append(group)
                
                if falseUserGroups:
                    Utils._error(self, f"Unknown user group(s) {falseUserGroups}")
                    return

            # Create role
            properties = dyno_client.structure[inventoryName]['properties']

            ppstring = '[' + ','.join(map(str.upper, propertiesPermissions)) + ']'
            props = '[\n'
            refProps = '[\n'
            for _, value in properties.items():
                if value["type"] == 'scalar':
                    props += f'{{ propertyId: {Utils._toGraphQL(value["propertyId"])}\n permissions: {ppstring} }}\n'
                elif value["type"] == 'reference':
                    refProps += f'{{ propertyId: {Utils._toGraphQL(value["propertyId"])}\n inventoryId: {Utils._toGraphQL(value["inventoryId"])}\n propertyPermissions: {ppstring}\n inventoryPermissions: [NONE]\n properties: []\n referencedProperties: []\n }}'
            props += ']'
            refProps += ']'
            
            graphQLString= f'''
            mutation AddRole($roleName: String!, $inventoryId: String!, $inventoryPermissions: [ObjectPermission!]!) {{ 
                addRole (input: {{
                    name: $roleName
                    rootInventoryPermission: {{
                        inventoryId: $inventoryId
                        inventoryPermissions: $inventoryPermissions
                        properties: {props}
                        referencedProperties: {refProps}
                        }}
                    }})
                    {{
                    id
                }}
            }}
            '''
            params = {
                "roleName": roleName,
                "inventoryId": dyno_client.structure[inventoryName]['inventoryId'],
                "inventoryPermissions": list(map(str.upper, objectPermissions)),
            }
            
            result = Utils._executeGraphQL(self, graphQLString, correlationId, params=params)
            if result == None: return

            # if result['addRole']['errors']:
            #     Utils._listGraphQlErrors(result, 'createInventory')
            #     return

            logger.info(f"Role {roleName} created.")

            roleId = result['addRole']['id']

            # Create rules
            if userGroups != None:
                for groupname in userGroups:
                    groupId = dfUserGroups.set_index('name').to_dict(orient='index')[groupname]['id']
                    createRuleGqlString= f'''
                    mutation AddRule($roleId: String!, $groupId: String!) {{
                        addRule (input: {{
                            roleId: $roleId
                            serviceAccountOrGroupId: $groupId
                            accountReferenceType: GROUP
                            filter: ""
                            filterFormat: EXPRESSION
                            }})
                            {{
                            ruleId
                        }}
                    }}
                    '''
                    result = Utils._executeGraphQL(self, createRuleGqlString, correlationId, params={"roleId": roleId, "groupId": groupId})
                    if result != None:
                        logger.info(f"Rule for {roleName} and user group {group} created.")
                    else:
                        logger.error(f"Rule for {roleName} and user group {group} could not be created.")

            return

    def deleteRole(self, role:str) -> None:
        """
        Deletes a role and all related rules.
        """

        correlationId = str(uuid4())
        with logger.contextualize(correlation_id=correlationId):

            # Get Ids of roles and rules
            roles = self.roles().set_index('name')
            roleId = roles.loc[role,'id']

            rules = self.rules()
            rules = rules.set_index('role.name')
            try:
                ruleIds = rules.loc[role,'id']
            except:
                ruleIds = []
            if not isinstance(ruleIds, str):
                ruleIds = list(ruleIds)
            else:
                ruleIds = [ruleIds]

            # First delete rules
            if ruleIds:
                deleteRuleGraphQLString = f'''
                mutation deleteRule($ruleId: String!) {{
                    removeRule(input: {{
                        ruleId: $ruleId
                    }}) {{
                        ruleId
                    }}
                }}
                '''
                for ruleId in ruleIds:
                    result = Utils._executeGraphQL(self, deleteRuleGraphQLString, correlationId, {"ruleId": ruleId})
                    if result != None:
                        logger.info(f"Rule {ruleId} of role {role} with id {ruleId} has been deleted.")
                    else:
                        Utils._error(self, f"Rule {ruleId} of role {roleId} could not be deleted.")
                        return

            # After all rules have been deleted, delete the role
            deleteRoleGraphQLString = f'''
            mutation deleteRole($roleId: String!) {{
                removeRole(input: {{
                    roleId: $roleId
                }}) {{
                    roleId
                }}
            }}
            '''
            result = Utils._executeGraphQL(self, deleteRoleGraphQLString, correlationId, {"roleId": roleId})
            if result != None:
                logger.info(f"Role {role} with id {roleId} has been deleted.")
            else:
                Utils._error(self, f"Role {roleId} could not be deleted.")
            
            return

    def addRule(self, role:str, userGroupOrServiceAccount:str, accountType:Literal["group", "service_account"]="group"):
        """
        Adds a rule connecting role with usergroup.
        """

        roleId = self.roles(fields=['id'], where=f'name eq "{role}"')['id'][0]
        if accountType == "group":
            serviceAccountOrGroupId = self.userGroups(fields=['id'], where=f'name eq "{userGroupOrServiceAccount}"')['id'][0]
        else:
            serviceAccountOrGroupId = self.serviceAccounts(fields=['id'], where=f'name eq "{userGroupOrServiceAccount}"')['id'][0]

        graphqlString= f'''
        mutation AddRule($roleId: String!, $serviceAccountOrGroupId: String!, $accountReferenceType: AccountReferenceTypeEnum!) {{
            addRule (input: {{
                roleId: $roleId
                serviceAccountOrGroupId: $serviceAccountOrGroupId
                accountReferenceType: $accountReferenceType
                filter: ""
                filterFormat: EXPRESSION
                }})
                {{
                ruleId
            }}
        }}
        '''
        logger.debug(graphqlString)
        params = {
            "roleId": roleId,
            "serviceAccountOrGroupId": serviceAccountOrGroupId,
            "accountReferenceType": "GROUP" if accountType == "group" else "SERVICE_ACCOUNT"
        }
        result = Utils._executeGraphQL(self, graphqlString, params=params)
        if result != None:
            logger.info(f"Rule for {role} and user group {userGroupOrServiceAccount} created.")
        else:
            Utils._error(self, f"Rule could not be created.")

        return



from uuid import uuid4
import pandas as pd
from loguru import logger

from .utils.ut import Utils
from .utils.ut_autprog import AutProgUtils

class Automation():
    accessToken = ''

    def __init__(self, accessToken:str, endpoint:str, client:object) -> None:
        
        self.raiseException = client.raiseException
        self.defaults = client.defaults

        self.accessToken = accessToken
        self.endpoint = endpoint
        self.proxies = client.proxies

    def getVersion(self):
        """
        Returns name and version of the responsible micro service
        """

        return Utils._getServiceVersion(self, 'automation')

    def workflows(self) -> pd.DataFrame:
        """Returns a DataFrame of all Workflows"""

        graphQLString = f'''query workflows {{
            workflows {{
                id
                name
                description
                }}
            }}
            '''

        result = Utils._executeGraphQL(self, graphQLString)
        if result == None: return

        df = pd.json_normalize(result['workflows'])
        return df

    def workflowInstances(self, workflowId:str=None, fromTimepoint:str=None, toTimepoint:str=None, 
        fields:list=None, where:str=None, showTasks=False) -> pd.DataFrame:
        """Shows Instances of a workflow. If workflowId=None, all Instances of all 
        workflows will be returned."""

        meta = ['id', 'name', 'businessKey', 'version', 'startTime', 'endTime', 'state']
        key = 'workflowInstances'

        if workflowId != None:
            _workflowId = f'workflowId: "{workflowId}"'
        else:
            _workflowId = ''

        if fromTimepoint != None:
            _fromTimepoint = f'from: "{fromTimepoint}"'
        else: 
            _fromTimepoint = ''

        if toTimepoint != None:
            _toTimepoint = f'from: "{toTimepoint}"'
        else: 
            _toTimepoint = ''

        if fields != None:
            if type(fields) != list:
                fields = [fields]
            _fields = Utils._queryFields(fields, recursive=True)   
        else:
            _fields =f'''
                id
                name
                businessKey
                version
                startTime
                endTime
                state
                variables {{
                    name
                    value
                    time
                }}''' 

        resolvedFilter = ''
        if where != None: 
            resolvedFilter = Utils._resolveWhereString(where)

        if showTasks != False:
            _tasks = f'''tasks {{
                            id
                            topic
                            workerId
                            timestamp
                            state
                            retries
                            errorMessage
                        }}'''
        else:
            _tasks = ''

        graphQLString = f'''query Instances {{
            {key}({_workflowId}, {_fromTimepoint}, {_toTimepoint}, {resolvedFilter}, all:true) {{
                {_fields}
                {_tasks}
                }}
            }}
            '''
     
        result = Utils._executeGraphQL(self, graphQLString)
        if result == None: return

        if showTasks != False:
            df = pd.json_normalize(result[key], meta=meta, record_path=['tasks'], record_prefix='task.', errors='ignore')
            if 'startTime' in df.columns:
                df = df.sort_values(by='startTime', ascending=False)
        else:
            df = pd.json_normalize(result[key])
            if 'startTime' in df.columns:
                df = df.sort_values(by='startTime', ascending=False)
        return df
    
    def createWorkflow(self, id, name, description:str=None):

        correlationId = str(uuid4())
        context_logger = logger.bind(correlation_id = correlationId)

        graphQLString = f'''mutation createWorkflow {{
            createWorkflow(
                input: {{
                    id: "{id}"
                    name: "{name}"
                    description: "{description}"
                }}
                ) {{
                    ...on CreateWorkflowError {{
                    message
                    }}
                    ... on WorkflowCreated {{
                        workflow {{
                            id
                        }}
                    }}
                }}
            }}
        '''
        
        result = Utils._executeGraphQL(self, graphQLString)
        if result == None: return

        context_logger.info(f"New workflow {id} created.")

        return result
                        
    def deployWorkflow(self, workflowId:str, filePath:str) -> None:
        """Deploys a Camunda XML to an existing workflow"""

        correlationId = str(uuid4())
        context_logger = logger.bind(correlation_id = correlationId)

        fileContent = Utils._encodeBase64(filePath)
        context_logger.debug(f"fileContent: {fileContent[:10]}")

        graphQLString = f'''mutation deployWorkflow {{
            deployWorkflow(
                input: {{
                    fileContentBase64: "{fileContent}"
                    workflowId: "{workflowId}"
                }}
            ) {{
                ... on DeployWorkflowError {{
                    message
                }}
                ... on InvalidWorkflowProcessId {{
                    processId
                    workflowId
                    message
                }}
                ... on WorkflowDeployed {{
                    version
                }}
            }}
        }}
        '''

        result = Utils._executeGraphQL(self, graphQLString)
        if result == None: return

        context_logger.info(f"Workflow '{workflowId}' deployed.")
        return result

    def startWorkflow(self, workflowId:str, businessKey:str, inputVariables:dict=None):
        """Starts a workflow"""

        correlationId = str(uuid4())
        context_logger = logger.bind(correlation_id = correlationId)

        if inputVariables == None:
            _vars = ''
        else:
            _vars = AutProgUtils._varsToString(inputVariables, 'input')

        graphQLString = f'''
            mutation ExecuteWF {{
                startWorkflow(input: {{ 
                    businessKey: "{businessKey}"
                    workflowId: "{workflowId}" 
                    {_vars}
                    }}
                ) {{
                    ... on ProcessDefinitionNotFound {{
                        workflowId
                        message
                        }}
                    ... on StartWorkflowError {{
                            message
                            }}
                    ... on WorkflowStarted {{
                        workflowInstanceId
                        }}
                    }}
                }}
            '''

        result = Utils._executeGraphQL(self, graphQLString)
        if result == None: return

        context_logger.info(f"Workflow {workflowId} started.")

        return result

    def deleteWorkflow(self, workflowId:str):
        """Deletes a workflow"""

        correlationId = str(uuid4())
        context_logger = logger.bind(correlation_id = correlationId)

        graphQLString = f'''mutation deleteWorkflow {{
            deleteWorkflow (id: "{workflowId}")
            {{
                ... on DeleteWorkflowError {{
                    message
                    }}
                ...on WorkflowDeleted {{
                    success
                    }}
                ... on WorkflowNotFound {{
                    workflowId
                    message
                    }}
                
                }}
            }}
        '''

        result = Utils._executeGraphQL(self, graphQLString)
        context_logger.info(f"Workflow {workflowId} deleted.")
        return result

    def terminateWorkflowInstance(self, workflowInstanceId):
        """Terminates a workflow instance"""

        correlationId = str(uuid4())
        context_logger = logger.bind(correlation_id = correlationId)

        graphQLString = f'''mutation terminateWorkflowInstance {{
            terminateWorkflowInstance(
                workflowInstanceId:"{workflowInstanceId}") {{
                ...on TerminateWorkflowInstanceError {{
                    message
                    }}
                ...on WorkflowInstanceNotFound {{
                    workflowInstanceId
                    message
                    }}
                ...on WorkflowInstanceTerminated {{
                    success
                    }}
                }}
            }}
        '''

        result = Utils._executeGraphQL(self, graphQLString)
        context_logger.info(f"Workflow instance {workflowInstanceId} started.")
        return result

    def updateWorkflow(self, workflowId:str, name:str=None, description:str=None):
        """Updates a workflow (name and description can be changed)"""

        correlationId = str(uuid4())
        context_logger = logger.bind(correlation_id = correlationId)

        name = Utils._argNone('name', name)
        description = Utils._argNone('description', description)

        key = 'updateWorkflow'
        graphQLString = f'''mutation updateWorkflow {{
            {key}(workflowId: "{workflowId}", properties: {{
                {description}
                {name}
                }}) {{
                    ... on UpdateWorkflowError {{
                    message
                    }}
                    ... on WorkflowNotFound {{
                    workflowId
                    message
                    }}
                    ... on WorkflowUpdated {{
                    workflow {{
                        id
                        name
                        description
                        }}
                    }}
                }}
            }}
            '''

        result = Utils._executeGraphQL(self, graphQLString)
        context_logger.info(f"Workflow {workflowId} updated.")
        return result

    def retryTask(self, externalTaskId):
        key = 'retryTask'

        correlationId = str(uuid4())
        context_logger = logger.bind(correlation_id = correlationId)

        graphQLString = f'''mutation retryTask {{
            {key}(externalTaskId: "{externalTaskId}") {{
                ... on RetryTaskError {{
                message
                }}
                ...on TaskNotFound {{
                message
                }}
                ... on TaskRetried {{
                success
                }}
            }}
        }}
        '''

        result = Utils._executeGraphQL(self, graphQLString)
        context_logger.info(f"Task {externalTaskId} retried.")
        return result



from setuptools import setup, find_packages
import pathlib

here = pathlib.Path(__file__).parent.resolve()
long_description = (here / 'README.md').read_text(encoding='utf-8')
version = ''
with open(here / 'seven2one/__init__.py') as fp:
      while version == '':
            line = fp.readline()
            if (line.startswith("__version__")):
                  version = line.replace("__version__", "").replace("=", "").replace('"', "").replace("'", "").strip()
            if not line:
                  break

setup(name='seven2one',
      version=version,
      description='Functions to interact with the Seven2one TechStack',
      long_description=long_description,
      long_description_content_type="text/markdown",
      url='http://www.seven2one.de',
      author='Seven2one Informationssysteme GmbH',
      author_email='info@seven2one.de',
      license='MIT',
      packages=['seven2one', 'seven2one.utils', 'seven2one.logging_loki'],
      include_package_data=True,
      install_requires=[
            'pandas>=1.4.2,<2.0.0', 'gql==3.0.0', 'pytz', 'tzlocal', 'pyperclip', 'loguru', 'requests', 'requests_toolbelt', 'rfc3339'
            ],
      classifiers =[
            'Development Status :: 3 - Alpha',
            'Natural Language :: English',
            'Operating System :: OS Independent',
            ],
      python_requires='>=3.8',
      zip_safe=False)


import atexit
from loguru import logger
from ..logging_loki import LokiQueueHandler, emitter
from . import __version__
from multiprocessing import Queue

class LogUtils():
    def _init_logging(
        url: str, 
        access_token: str,
        log_level: str, 
        sessionId:str, 
        clientId: str="S2O.TechStack.Python"):

        emitter.LokiEmitter.level_tag = "level"
        handler = LokiQueueHandler(
            Queue(),
            url=url,
            tags={"client": clientId},
            version="1",
            token=access_token
        )

        # logger.remove()
        # logger.add(sys.stderr, level=log_level)
        logger.add(handler, level=log_level, serialize=True, backtrace=True, diagnose=True)
        logger.configure(extra={
            "version": __version__,
            "session_id": sessionId,
            })

        def _teardown_logging(handler):
            handler.listener.stop()

        atexit.register(_teardown_logging, handler)
